!=========================================================================
! Copyright (C) GemTalk Systems 1986-2019.  All Rights Reserved.
!
! $Id: gsbitmap.gs 38462 2016-01-13 04:33:12Z lalmarod $
!
! Superclass Hierarchy:
!   GsBitmap, Object.
!
!=========================================================================

expectvalue %String
run
Globals at: #GsBitmap ifAbsent:[
  Object _newKernelSubclass: 'GsBitmap'
    subclassOf: Object
    instVarNames: #( hiddenSetId )
    classVars: #()
    classInstVars: #()
    poolDictionaries: #()
    inDictionary: Globals
    options:  #( #instancesNonPersistent )
    reservedOop: 1949 .
  ^ ' created GsBitmap'
].
^ 'exists'
%
run
| o |
(o := GsBitmap asOop) == 249345 ifFalse:[ ^ o ].
^ true
%

removeallmethods GsBitmap
removeallclassmethods GsBitmap
set class GsBitmap

category: 'For Documentation Installation only'
classmethod: GsBitmap
installDocumentation

self comment:

'GsBitmap is a class that implements an in memory bitmap, i.e. a transient
 bitmap that cannot be made persistent. A bitmap is a sparse data structure
 that logically implements a bit array. Each bit in the bitmap represents
 the objectId of a committed non-special object.
 Access to the hiddenSets in the system is implemented by creating 
 a GsBitmap that references a particular hiddenSet (see newForHiddenSet:).  

 Most of the hiddenSets in the system are read only and generate an error
 if the operation is not allowed.  See the code for the hiddenSetSpecifiers 
 method and note the last mutable hiddenSet.  HiddenSets after this point in 
 the  array generate errors if an attempt is made to modify them and the user 
 is not SystemUser.  The NotifySet, PureExportSet, and GciTrackedObjs
 have interfaces defined in System class (categories are NotifySet and
 GcISets) and these interfaces should be used to guarantee correct behavior.
 Updates to these 3 sets generate errors when attempted from GsBitmap.

 An instance of GsBitMap uses C Heap memory to store the bit array. 
 The C Heap memory associated with an instance of GsBitmap is automatically 
 freed when the instance is garbage collected.

 Do not retain an instance of GsBitmap from one session to another.
 Instances of GsBitmap are intended to exist only within a given GemStone
 session. 

 Instances may not be committed.

 The objects referenced in a GsBitmap are not protected from garbage collection.
'
%

category: 'Hidden Set Support'
classmethod: GsBitmap
hiddenSetSpecifiers

"Returns a list of the hidden set specifiers.

**************************************************************************************************
***** DO NOT CHANGE THE ORDER OF THE ELEMENTS IN THE LIST WITHOUT EDITING CODE IN gsbitmap.c *****
**************************************************************************************************
"

^#(
 #ListInstances
 #SaveNewPomObjs "See method #_enableTraceNewPomObjs"
 #ObjectsRead    "See method #_enableTraceObjectsRead"
 #SaveWriteSetUnion
 #CommitReleaseLocksSet
 #CommitOrAbortReleaseLocksSet
 #StrongReadSet  
 #GcCandidates "last mutable hiddenSet"

 #RcReadSet
 #DepMapWriteSet
 #PomWriteSet    "Empty except after flush for commit, so only useful
		       after a transaction conflict."
 #RemovedDepMapEntries
 #SaveDepMapChangedUnion
 #ReadWriteConflicts  "StrongRead-Write conflicts "
 #WriteWriteConflicts
 #WriteDependencyConflicts
 #WriteReadLockConflicts
 #WriteWriteLockConflicts
 #AllocatedGciOops
 #ExportedDirtyObjs
 #TrackedDirtyObjs
 #ReferencedSet
 #WriteLockWriteSubset
 #NewDataPages
 #PreviousWsUnion
 #PureExportSet
 #GciTrackedObjs
 #NotifySet          
)
%

category: 'Private'
method: GsBitmap
_1ArgPrim: opCode

"opCode     function
   1        _basicNew
   2        removeAll
   3        size
   4        asArray
   5        asArrayOfOops
   6        copy
   7        sizeInPages
   8        peakLast
   9        referencedObjects
   10       _asArrayOfBits
"

<primitive: 1026>
opCode _validateClass: SmallInteger.
self _primitiveFailed: #_1ArgPrim: args: { opCode }
%

category: 'Instance Creation'
classmethod: GsBitmap
basicNew

"disallowed"

self shouldNotImplement: #basicNew
%

category: 'Private'
method: GsBitmap
_basicNew

"creates an instance registered with VM for finalization of cData"

<primitive: 674>
self _primitiveFailed: #_basicNew
%

category: 'Instance Creation'
classmethod: GsBitmap
new

"Returns a new empty GsBitmap" 

^ self _basicNew _1ArgPrim: 1
%

category: 'Instance Creation'
classmethod: GsBitmap
with: aValue

"Returns an instance of the receiver containing the argument" 

| inst |
inst := self new.
inst add: aValue.
^inst
%

category: 'Instance Creation'
classmethod: GsBitmap
with: aValue with: val2

"Returns an instance of the receiver containing the arguments" 

| inst |
inst := self new.
inst add: aValue; add: val2.
^inst
%

category: 'Instance Creation'
classmethod: GsBitmap
with: aValue with: val2 with: val3

"Returns an instance of the receiver containing the arguments" 
| inst |
inst := self new.
inst add: aValue; add: val2; add: val3.
^inst
%

category: 'Instance Creation'
classmethod: GsBitmap
with: aValue with: val2 with: val3 with: val4

"Returns an instance of the receiver containing the arguments" 
| inst |
inst := self new.
inst add: aValue; add: val2; add: val3; add: val4.
^inst
%

category: 'Instance Creation'
classmethod: GsBitmap
withAll: aCollection

"Returns an instance of the receiver containing the elements of the argument."

| inst | 
inst := GsBitmap new.
aCollection accompaniedBy: inst do: [ :me :each | me add: each ].
^ inst.
%

category: 'Private'
method: GsBitmap
_hiddenSetId

^hiddenSetId
%

category: 'Private'
method: GsBitmap
_hiddenSetId: aSmallInt

hiddenSetId := aSmallInt
%

category: 'Instance Creation'
classmethod: GsBitmap
newForHiddenSet: hiddenSetSpecifier

"Returns a GsBitmap that references the specified hiddenSet."

| gsBitmap idx | 
gsBitmap := GsBitmap new.
idx := GsBitmap hiddenSetSpecifiers indexOfIdentical: hiddenSetSpecifier.
idx < 1 ifTrue: [ ^ hiddenSetSpecifier _error: #rtErrInvalidArgument 
                              args:{ 'not a valid hiddenSetSpecifier' }] .
gsBitmap _hiddenSetId: idx.
^gsBitmap
%

category: 'Accessing'
method: GsBitmap
hiddenSetName

"Returns the name as a symbol that corresponds to this instance.
 If there is no hidden set reference then it returns the symbol #NotAHiddenSet."

 hiddenSetId == 0 ifTrue: [ ^ #NotAHiddenSet ]
                  ifFalse: [ ^ (GsBitmap hiddenSetSpecifiers) at: hiddenSetId ]
%

category: 'Removing'
method: GsBitmap
removeAll

"Clears all of the bits in the receiver."

^self _1ArgPrim: 2
%

category: 'Testing'
method: GsBitmap
size 

"Returns the number of bits that are set in the GsBitmap."

^self _1ArgPrim: 3
%

category: 'Accessing'
method: GsBitmap
asArray

"Returns the bits that are set in the GsBitmap as an Array of objects.
 The results is an object in temporary object memory."

^self _1ArgPrim: 4
%

category: 'Private'
method: GsBitmap
_asArrayOfOops

"Returns the bits that are set in the GsBitmap as an Array of SmallIntegers
 which correspoind to the oops of the persistent objects.
 This method can be used to return the results of a large GsBitmap in 
 an array without faulting all of the objects into memory like asArray does.
 The oops can be used in the method _objectForOop: to convert the value to an object."

^self _1ArgPrim: 5
%
category: 'Private'
method: GsBitmap
_asArrayOfBits

"Returns the bits that are set in the GsBitmap as an Array of Integers
 which correspoind to the raw bit numbers in the receiver."

^ self _1ArgPrim: 10
%

category: 'Accessing'
method: GsBitmap
copy

"Returns a new GsBitmap that is a copy of the receiver."

^self _1ArgPrim: 6
%

category: 'Accessing'
method: GsBitmap
sizeInPages

"Returns the number of 16 K pages used to implement the receiver."

^self _1ArgPrim: 7
%

category: 'Accessing'
method: GsBitmap
peakLast

"Returns the last object in the GsBitmap"

^self _1ArgPrim: 8
%

category: 'Enumerating'
method: GsBitmap
referencedObjects

"Answer a new GsBitmap that contains the committed, non-special objects
referenced by the objects contained in the receiver.  For each object for
which the current user has read authorization, the objects that are directly
referenced stucturally by that object (from named instance variables, indexed
or NSC contents, and dynamic instance variables) are included."

^self _1ArgPrim: 9
%

category: 'Private'
method: GsBitmap
_2ArgPrim: opCode with: arg2

" Functions with the receiver readOnly should be first
 opCode     function      
   1        includes:  
   2        writeToFile: 
   3        writeToFileInPageOrder:
   4        equals:
   5        firstObjectThatReferences:
     First mutable opcode
   6        add:         primitive returns the previous state
   7        addAll:  
   8        remove:      primitive returns the previous state
   9        removeAll: 
  10        union
  11        difference
  12        intersect
  13        truncateToSize:
  14        removeFirst

  20        fileInfo:
  21        readFromFile: 
  22        auditFile:
  23        _addBits:
"

<primitive: 1027>
opCode _validateClass: SmallInteger.
self _primitiveFailed: #_2ArgPrim:with: args: { opCode . arg2 }
%

category: 'Testing'
method: GsBitmap
includes: anObject

"Returns true if anObject is in the GsBitmap."

^self _2ArgPrim: 1 with: anObject
%

category: 'Private'
method: GsBitmap
_truncateToSize: newSize

"Truncate receiver by removing objects from the end of the set until it reaches a size
 of newSize.  Objects are removed in order, going from highest to lowest object ID.   
 
 Returns the number of objects removed."
 
^self _2ArgPrim: 13 with: newSize
%

category: 'Private'
method: GsBitmap
_removeFirst: count 

"Remove the first count objects from the receiver.  Objects are removed from
 the beginning, going from lowest to highest object ID. 
 
 Returns the number of objects removed from the hidden set."
 
^self _2ArgPrim: 14 with: count
%

category: 'Removing'
method: GsBitmap
removeFirst: count 

"Remove the first count objects from the receiver and return them in a new GsBitmap.
 Objects are removed from the beginning, going from lowest to highest object ID.  
"

| copy |
copy := self copy.
self _removeFirst: count.
^copy removeAll: self
%

category: 'Removing'
method: GsBitmap
removeLast: count 

"Remove the last count objects from the receiver and return them in a new GsBitmap.
 Objects are removed from the end, going from higest to the lowest object ID.  
"

| copy |
copy := self copy.
self _truncateToSize: (self size - count).
^copy removeAll: self
%

category: 'Testing'
method: GsBitmap
isEmpty

"Returns true there are no entries in the GsBitmap."

^self size == 0
%

category 'File Operations'
method
writeToFile: aString

"Writes the contents of the receiver to the specified file.
 The file created is approximately 5 bytes for each object plus 24 bytes for 
 the file header.

 If the file exists an error is returned.

 Returns the number of objectIds written to the file.
"

^self _2ArgPrim: 2 with: aString
%

category 'File Operations'
method
writeToFileInPageOrder: aString
"Writes the contents of the receiver to the specified file in pageOrder.

 Operations like instance migration which process a large volume of objects 
 often run significantly faster if the objects are processed in batches that
 are in page order rather than in object ID order.

 The file created is approximately 5 bytes for each object plus 24 bytes for 
 the file header.

 If the file exists an error is returned.

 Returns the number of objectIds written to the file.
"

^self _2ArgPrim: 3 with: aString
%

category: 'Testing'
method: GsBitmap
equals: aGsBitmap

"Returns true if the receiver and argument contain exactly the same elements."

^self _2ArgPrim: 4 with: aGsBitmap
%

category: 'Searching'
method: GsBitmap
firstObjectThatReferences: anObject
"Search the receiver to find and return the first object which references anObject. 
 Returns nil if no such object was found."

^self _2ArgPrim: 5 with: anObject
%

category: 'Adding'
method: GsBitmap
add: anObject

"Adds a non special object to the GsBitmap.  
 Returns anObject."

self _2ArgPrim: 6 with: anObject.
^anObject
%

category: 'Adding'
method: GsBitmap
addAll: anArrayOrGsBitmap

"Adds each element of the Array of non special objects or the contents of a GsBitmap 
 to the receiver. Returns an Integer, the number of bits added to the receiver."

(anArrayOrGsBitmap == nil) ifTrue: [ ^anArrayOrGsBitmap ].
^self _2ArgPrim: 7 with: anArrayOrGsBitmap
%

category: 'Private'
method:
_setBits: anArray

"Add the bits specified by the Integers in anArray to the receiver.
 Returns the number of bits added."

^self _2ArgPrim: 23 with: anArray
%
category: 'Private'
method:
_clearBits: anArray

"Clears the bits specified by the Integers in anArray from the receiver.
 Returns the number of bits cleared."

^self _2ArgPrim: 24 with: anArray
%


category: 'Removing'
method: GsBitmap
remove: anObject

"Removes anObject from the GsBitmap.
 If the object is not present in the GsBitmap it issues #objErrNotInColl,
 otherwise it returns anObject.
"
(self _2ArgPrim: 8 with: anObject) 
    ifFalse: [ self _error: #objErrNotInColl args: { anObject } ].
^anObject
%

category: 'Removing'
method: GsBitmap
remove: anObject ifAbsent: aBlock

"Removes anObject from the receiver.  If anObject is not in the
 receiver, this method evaluates aBlock and returns its value.  
 The argument aBlock must be a zero-argument block.
"

(self _2ArgPrim: 8 with: anObject)
    ifFalse: [ ^ aBlock value ].
^anObject
%

category: 'Removing'
method: GsBitmap
removeIfPresent: anObject

"Remove anObject from the receiver.  Returns nil if anObject is
 missing from the receiver, otherwise returns the argument.
"

(self _2ArgPrim: 8 with: anObject)
    ifFalse: [ ^ nil ].
^anObject
%

category: 'Removing'
method: GsBitmap
removeAll: anArrayOrGsBitmap

"Removes each element of the Array of non special objects or the contents of a GsBitmap 
 from the receiver.  Returns anArrayOrGsBitmap."

^self _2ArgPrim: 9 with: anArrayOrGsBitmap
%

category: 'Private'
method: GsBitmap
_3ArgPrim: opCode with: arg2 with: arg3

" Functions with the receiver readOnly should be first
 opCode     function      
   1        enumerateWithLimit:startingAfter:
   2        enumerateAsOopsWithLimit:startingAfter:
 "

<primitive: 1028>
opCode _validateClass: SmallInteger.
self _primitiveFailed: #_3ArgPrim:with:with: args: { opCode . arg2 . arg3}
%

category: 'Private'
method: GsBitmap
_enumerateWithLimit: maxResultSize startingAfter: anObject

"Returns an array containing the first maxResultSize bits as objects
 starting with the first one after the specified object.  
 To start the enumeration use the SmallInteger zero as anObject.
 To get the next batch use the last element of the array returned.

 This enumeration is non destructive, so no values are cleared.

 If the GsBitmap contains fewer than maxResultSize elements after the
 specified object or the maxResultSize has the value 0, then it returns 
 all of the elements after the specified object.
"

^self _3ArgPrim: 1 with: maxResultSize with: anObject
%

category: 'Private'
method: GsBitmap
_enumerateAsOopsWithLimit: maxResultSize startingAfter: anOop

"Returns an array containing the first maxResultSize bits as oops
 starting with the first one after the specified oop.
 To start the enumeration use the SmallInteger zero as anOop. 
 To get the next batch use the last element of the array returned.

 This enumeration is non destructive, so no values are cleared.

 If the GsBitmap contains fewer than maxResultSize elements after the
 specified object or the maxResultSize has the value 0, then it returns 
 all of the elements after the specified object.
"

^self _3ArgPrim: 2 with: maxResultSize with: anOop
%

category: 'Enumerating'
method: GsBitmap
do: aBlock

"Executes the one argument block aBlock for each object in 
 in the receiver.  Does a non destructive enumerate of the GsBitmap. 
 Returns the number of objects enumerated."

 | count arr |
 count := 0.
 arr := Array new.
 arr add: 0.
 [ | n |
   arr := self _enumerateWithLimit: 2000 startingAfter: (arr last).
   n := arr size .
   count := count + n .
   1 to: n do:[:j | aBlock value: (arr at: j) ] .
   n ~~ 0 . 
 ] whileTrue.
 ^ count
%

category: 'Private'
method: GsBitmap
_doAsOops: aBlock

"Executes the one argument block aBlock for each oop in 
 in the receiver.  Does a non destructive enumerate of the GsBitmap 
 with the argument to the block being anOop. 
 Returns the number of objects enumerated."

 | count arr |
 count := 0.
 arr := Array new.
 arr add: 0.
 [ | n |
   arr := self _enumerateAsOopsWithLimit: 2000 startingAfter: (arr last).
   n := arr size .
   count := count + n .
   1 to: n do:[:j | aBlock value: (arr at: j) ] .
   n ~~ 0 . 
 ] whileTrue.
 ^ count
%


category: 'Set Arithmetic'
method: GsBitmap
+ aGsBitmap

"Returns a new GsBitmap with the elements that are in the receiver or aGsBitmap.
 Same as union:."

^self _2ArgPrim: 10 with: aGsBitmap
%

category: 'Set Arithmetic'
method: GsBitmap
union: aGsBitmap

"Returns a new GsBitmap with the elements that are in the receiver or aGsBitmap.
 Same as +."

^self _2ArgPrim: 10 with: aGsBitmap
%

category: 'Set Arithmetic'
method: GsBitmap
- aGsBitmap

"Returns a new GsBitmap with the elements in the receiver that are not in aGsBitmap.
 Same as difference:."

^self _2ArgPrim: 11 with: aGsBitmap
%

category: 'Set Arithmetic'
method: GsBitmap
difference: aGsBitmap

"Returns a new GsBitmap with the elements in the receiver that are not in aGsBitmap.
 Same as -."

^self _2ArgPrim: 11 with: aGsBitmap
%

category: 'Set Arithmetic'
method: GsBitmap
* aGsBitmap

"Returns a new GsBitmap with the elements that are in both the receiver and in aGsBitmap.
 Same as inersect:."

^self _2ArgPrim: 12 with: aGsBitmap
%

category: 'Set Arithmetic'
method: GsBitmap
intersect: aGsBitmap

"Returns a new GsBitmap with the elements that are in both the receiver and in aGsBitmap.
 Same as *."

^self _2ArgPrim: 12 with: aGsBitmap
%

category: 'File Operations'
classmethod: GsBitmap
fileInfo: bitmapFilePath

" Returns an array which contains the following values:
  1.  The number of oops in the file
  2.  True if the file was written in page order.
  3.  The number of valid oops.
  4.  The number of invalid oops (oops that are not allocated or in the 
      process of being garbage collected).

  Raises an error if the file could not be opened for reading, the file
  is corrupt, or aString is not an instance of String.
"

^(GsBitmap new) _2ArgPrim: 20 with: bitmapFilePath
%

category: 'Enumerating'
method: GsBitmap
readFromFile: aString

"Reads the contents of the file and adds the oops in the file to the 
 receiver.  If the file doesn't exist or is not in the GsBitmap format, 
 then this method returns an error. 

 Subsets of the file can be read into a GsBitmap with 
         readFromFile:withLimit:startingAt:.

 ObjectIds that are in the freelist or in the deadObjs or possibleDead sets 
 are not included. 

 Returns the number of objects that were not included.
"

^self _2ArgPrim: 21 with: aString
%

category: 'Private'
method
_auditFile: aString
 "Reads the file. "

 ^ self _2ArgPrim: 22 with: aString
%

category: 'File Operations'
method: GsBitmap
readFromFile: aString withLimit: maxResultSize startingAt: startIndex

"Adds the contents of the specified file indexed by the subrange: 
 (startingAt, startingAt + maxResultSize).  The resulting GsBitmap 
 may contain fewer than maxResultSize elements if the value of
 startingAt + maxResultSize is an index that is larger than the number of
 oops in the file or if some of the ObjectIds represented are no longer 
 live objects, i.e., they were found to be in the freeList or the 
 deadNotReclaimed or possible dead sets.

 If the file doesn't exist or is not in the GsBitmap format, 
 then this method returns an error. 

 The first object in the file has an index of 1.  It is an error if 
 startIndex is less than 1.

 Returns the number of objects read.
"

<primitive: 1029>
aString _validateClass: String.
maxResultSize _validateClass: SmallInteger.
startIndex _validateClass: SmallInteger.
self _primitiveFailed: #readFromFile:withLimit:startingAt: args: { aString . maxResultSize . startIndex }
%

category: 'Formatting'
method:
printString

"Returns a String with a displayable representation of the receiver."

^ self printStringWithMaxSize: 1000
%


category: 'Formatting'
method: GsBitmap
printStringWithMaxSize: n

"Returns the contents of the GsBitmap as a string with oops of the 
 objects in the bitmap arranged 8 per line.
"
| count result bmSize |

count := 0.
result := String new.
bmSize := self size.
result add: ('GsBitmap contains ' , bmSize asString , ' objects. ' , Character lf).
(bmSize == 0) ifTrue: [ ^ result ].
result add: ('1. ').
self _doAsOops: [ :oop | 
  count := count + 1.
  result add: (' ' , oop asString , ' ').
  ((count rem: 8) = 0 and: [count < bmSize]) ifTrue: [ 
     result add: (' ' , Character lf , (count + 1) asString , '. ').
     (result size > n and: [n ~= 0]) ifTrue: [ 
       result add: (' ........' , Character lf).
       result add: ( bmSize asString , '. ' , self peekLast asString, Character lf).
       ^ result
     ]
  ]
].
^result
%

category: 'Formatting'
method: GsBitmap
printOn: aStream

"Puts a displayable representation of the receiver on aStream."

aStream nextPutAll: (self printStringWithMaxSize: 0).
%

category: 'Peeking'
method: GsBitmap
peek

"Answer the first element in the receiver without removing it, or nil if the receiver is empty."
^ self isEmpty ifTrue:[nil] ifFalse:[ (self peekCount: 1) at: 1]
%

category: 'Peeking'
method: GsBitmap
peekLast 

"Answer the last in the receiver without removing it, or nil if the receiver is empty."
^ self isEmpty ifTrue:[nil] ifFalse:[ self _1ArgPrim: 8 ]
%

category: 'Peeking'
method: GsBitmap
peekCount: aCount

"Fetches elements from the receiver without removing them.  Returns an Array that contains
 the minimum of aCount or the receiver's size entries."

^self _enumerateWithLimit: aCount startingAfter: 0
%

category: 'Set Arithmetic'
method: GsBitmap
intersectSize: aGsBitmap

"Returns a SmallInteger which is the number of elements present in both the receiver and aGsBitmap."

| result newBm |
newBm := self * aGsBitmap .
result := newBm size.
newBm removeAll .
^ result
%

category: 'Removing'
method: GsBitmap
removeCount: maxToRemove
"Removes entries from the receiver, and returns an Array that contains the
 minimum of maxToRemove or the receiver's size entries."

|result|
result := self peekCount: maxToRemove .
self removeAll: result .
^result
%

category: 'Removing'
method: GsBitmap
removeNext
"Removes the first element from the receiver and returns it.  Returns nil if the receiver is empty."

^ self isEmpty ifTrue:[nil] ifFalse:[ (self removeCount: 1) at: 1]
%
category: 'Converting'
method: GsBitmap
asGsBitmap

^ self
%

